from .pager import *
import json
import pandas as pd
import numbers

class group:

    def __init__(self, api, data):        
        """
        Internal use only: initialize sample serie.
        """
        if ((data is not None) & (type(data) == dict) & 
            ("groupID" in data.keys()) & 
            ("name" in data.keys())
           ):
            self.__api = api
            self.__groupID = data["groupID"]
            self.__name = data["name"]
            self.__title = "Group '"+self.__name+"' ("+str(self.__groupID)+")"
            self.__data = data
        else:
            raise Exception("no (valid) group data") 
        
    def __repr__(self):
        """
        Internal use only: description group object.
        """ 
        return(self.__title)
        
    def name(self):
        """
        Get the name of the group.
        """
        return(self.__name)
        
    def id(self):
        """
        Get the id of the group.
        """
        return(self.__groupID)    
        
    def data(self):
        """
        Get the data describing the group.
        """
        return(self.__data)    
        
    


from .pager import *
import json
import pandas as pd

class sample_type_meta:

    def __init__(self, api, data):        
        if ((data is not None) & (type(data) == dict) & 
            ("sampleTypeMetaID" in data.keys()) & 
            ("key" in data.keys())
           ):
            self.__api = api
            self.__sampleTypeMetaID = data["sampleTypeMetaID"]
            self.__key = data["key"]
            self.__title = "Sample type meta '"+self.__key+"' ("+str(self.__sampleTypeMetaID)+")"
            self.__data = data
        else:
            raise Exception("no (valid) sampleTypeMeta data") 
        
    def __repr__(self):
        return(self.__title)
        
    def key(self):
        return(self.__key)
        
    def id(self):
        return(self.__sampleTypeMetaID)    
        
    def data(self):
        return(self.__data)   
           
    


import pandas as pd
import math
import ipywidgets as widgets
import warnings
from IPython.display import display, clear_output
from pandas.io.json import json_normalize


class pager:

    def __init__(self, api, title, location, request, index, sort, records, item_handler=None):        
        """
        Internal use only: initialize pager object: general object to browse items
        
        Parameters
        ----------
        api: api
            The api object
        title: str
            Title for the set of items
        location: str
            Prefix for REST api call url
        request: dict
            Parameters to add to REST api call url
        index: str / list
            Fields from response to be used as index.
            Also used as parameters for optional item_handler.
        sort: str
            Field used to sort response
        records: int
            Number of items on each page when browsing the response.
        item_handler: method, optional
            Method to be used to get an item.            
        """
        #store
        self.__api = api
        self.__title = title
        self.__location = location
        self.__request = request
        self.__index = index
        self.__sort = sort
        self.__records = records
        self.__item_handler = item_handler
        self.__page = 0
        self._reset()
        #get first page        
        self._set_page(self.__page, records) 
        #define first
        if len(self.__data)>0:
            self.__first = self.__data.index.values[0]
        else:
            self.__first = None     
            
    def _reset(self):
        """
        Internal use only: reset pager object
        """ 
        self.__page = 0
        self.__maxRecords = 0
        self.__totalRecords = 0
        self.__pages = 0
        self.__data = None
        
    def __repr__(self):
        """
        Internal use only: description pager object
        """ 
        description = self.__title
        if self.__pages>1:
            return(description+" ("+str(self.__totalRecords)+"x)")
        else:
            return(description)
        
    def _set_page(self, page, records):
        """
        Internal use only: change page
        """ 
        #define request
        request = self.__request
        request["$page"] = page
        request["$records"] = records
        request["$sort"] = self.__sort
        rp = self.__api.request(self.__location,"get",request)
        #check and get
        if rp==None:
            self._reset()
        elif type(rp) == dict:
            if "maxRecords" in rp.keys():
                self.__maxRecords = rp["maxRecords"]
            else:
                raise Exception("unexpected response, no maxRecords")
            if "totalRecords" in rp.keys():
                self.__totalRecords = rp["totalRecords"]
            else:
                raise Exception("unexpected response, no totalRecords")    
            if "currentPage" in rp.keys():
                self.__page = rp["currentPage"]
            else:
                raise Exception("unexpected response, no currentPage")
            self.__pages = math.ceil(self.__totalRecords/self.__maxRecords)  
            if "data" in rp.keys():                        
                self.__page = page
                self.__data = pd.DataFrame(json_normalize(rp["data"]))
                if len(self.__data)>0:
                    self.__data = self._filter_page(self.__data.set_index(self.__index))                
            else:
                raise Exception("unexpected response, no data")                    
        else:
            raise Exception("unexpected response, no dict")  
            
    def _filter_page(self, data):
        """
        Internal use only: filter page (flatten specific json)
        """ 
        dropList = set()
        numberTypes = {}
        for key in data.index:
            value = data.loc[key]
            for subkey in value.index:
                subvalue = value.loc[subkey]
                if isinstance(subvalue, list):
                    couldFindReplace=False
                    replace = {}
                    for item in subvalue:
                        if isinstance(item,dict):
                            couldFindReplace=True
                            if "key" in item:
                                for itemKey in item:
                                    if not(itemKey=="key"):
                                        newItemKey = str(item["key"])+"."+str(itemKey)
                                        if newItemKey in replace:
                                            replace[newItemKey].append(item[itemKey])
                                        else:
                                            replace[newItemKey] = [item[itemKey]]                                        
                            else:
                                for itemKey in item:
                                    if itemKey in replace:
                                        replace[itemKey].append(item[itemKey])
                                    else:
                                        replace[itemKey] = [item[itemKey]]                                                                
                    for replaceKey in replace.keys():
                        replaceValue = replace[replaceKey]
                        if len(replaceValue)==1:
                            replace[replaceKey] = replaceValue[0]
                        data.loc[key,str(subkey)+"."+str(replaceKey)] = str(replace[replaceKey])
                    if couldFindReplace:
                        data.loc[key,str(subkey)+".number"] = str(len(subvalue)) 
                        numberTypes[str(subkey)+".number"] = int
                        dropList.add(subkey)                        
        for numberType in numberTypes:
            data[numberType] = data[numberType].fillna(0).astype(numberTypes[numberType])        
        if len(dropList)>0:                       
            data = data.drop(columns=list(dropList))
        return(data)                    
                   
    
    def first(self, single=False):  
        """
        Get the first item (if exists), but raise an exception if `single`
        is set to True and the object describes multiple items.
        
        If an `item_handler` is defined, this will be
        used to return an object. Otherwise the index for the first
        item will be returned (usually the ID).
        """ 
        if not((self.__item_handler is None) | (self.__first==None)):
            if single & (self.__totalRecords>1):
                raise Exception("multiple entries found ("+str(self.__totalRecords)+")")
            if isinstance(self.__first, tuple):
                return self.__item_handler(*self.__first)
            else:         
                return self.__item_handler(self.__first)
        else:    
            return(self.__first)
            
    def number(self):
        """
        Get the number of items.
        """   
        return(self.__totalRecords)   
            
    def fields(self):
        """
        Get the available fields.
        """   
        return(self.__data.columns.values.tolist())   
            
    def get(self, *args):
        """
        Get the item for the provided index entry.
        Only available if an `item_handler` is defined.
        This will also return items that are not contained within the scope of this object!
        """   
        if self.__item_handler is None:
            raise Exception("method not available")
        else:
           return self.__item_handler(*args)    
            
    def all(self, fields=None, maximum=None):
        """
        Return all items in a single DataFrame.
        
        Parameters
        ----------
        fields: list
            A list of fields to limit the number of columns in the result
        """         
        if maximum is None:
            records =  100
        else:
            records = min(100, maximum)  
        recordCounter = 0                 
        page = 0
        dataSets = []
        while True:
            #define request
            request = self.__request
            request["$page"] = page
            request["$records"] = records
            request["$sort"] = self.__sort
            rp = self.__api.request(self.__location,"get",request)
            if rp==None:
                return None;
            elif type(rp) == dict:
                if "maxRecords" in rp.keys():
                    maxRecords = rp["maxRecords"]
                else:
                    raise Exception("unexpected response, no maxRecords")
                if "totalRecords" in rp.keys():
                    totalRecords = rp["totalRecords"]
                else:
                    raise Exception("unexpected response, no totalRecords")    
                if "currentPage" in rp.keys():
                    page = rp["currentPage"]
                else:
                    raise Exception("unexpected response, no currentPage")
                pages = math.ceil(totalRecords/maxRecords)  
            if "data" in rp.keys():                        
                data = pd.DataFrame(json_normalize(rp["data"]))                
                #only if data
                if len(data)>0:
                    data = data.set_index(self.__index) 
                    data = self._filter_page(data)
                    #filter columns
                    if not(fields==None):
                        intersectList = [x for x in fields if x in data.columns.values]
                        dropList = [x for x in data.columns.values if x not in intersectList]
                        data = data.drop(columns=dropList)[intersectList]
                    if maximum is None:    
                        dataSets.append(data)
                        recordCounter += len(data)
                    else:
                        if (len(data)+recordCounter) <= maximum:
                            dataSets.append(data)
                            recordCounter += len(data)
                        else:
                            dataSets.append(data.head(maximum-(len(data)+recordCounter)))
                            recordCounter += maximum-(len(data)+recordCounter)
                            break                                  
                else:
                    break                   
            else:
                raise Exception("unexpected response, no data") 
            if recordCounter>=totalRecords:
                break    
            page+=1
        return pd.concat(dataSets)       
                                    
    
    def show(self, fields=None, records=None):   
        """
        Create a browseable representation of the items.
        
        Parameters
        ----------
        fields: list, optional
            A list of fields to limit the number of columns in the result
        records: integer, optional
            The number of items to show on each page when browsing the result
        """            
        #data
        data = widgets.Output()  
        if records is None:
            records = self.__records  
        self._set_page(self.__page, records)                               
        #paging
        box_layout = widgets.Layout(
            display='flex', flex_flow='row', justify_content='space-between', 
            border='solid', width='100%')
        children = []
        header = widgets.Label()    
        button_previous = widgets.Button(description="previous", button_style="primary")                        
        button_next = widgets.Button(description="next", button_style="primary")                        
        if self.__pages>1:
            children = [button_previous, header, button_next]
        else:
            children = [widgets.Label(),header,widgets.Label()]
        box = widgets.Box(children=children, layout=box_layout)        
        #output
        display(widgets.VBox((box,data)))
            
        #set header content    
        def set_header_and_buttons(page):  
            if self.__pages>1:
                header.value=self.__title+" - page "+str(page+1)+" of "+str(self.__pages)
                if page==0:
                    button_previous.disabled=True
                    button_previous.button_style=""
                else:
                    button_previous.disabled=False
                    button_previous.button_style="primary"
                if page==(self.__pages-1):
                    button_next.disabled=True
                    button_next.button_style="" 
                else:
                    button_next.disabled=False
                    button_next.button_style="primary"    
            else:
                header.value=self.__title
                
        #define actions on buttons    
        def on_button_previous(b):
            if self.__page>0:
                set_header_and_buttons(self.__page-1) 
                button_previous.disabled=True
                button_next.disabled=True
                self._set_page(self.__page-1, records) 
                set_header_and_buttons(self.__page) 
                with data:
                    clear_output();
                    if not(fields==None):
                        rawData = self._filter_page(self.__data)
                        intersectList = [x for x in fields if x in rawData.columns.values]
                        dropList = [x for x in rawData.columns.values if x not in intersectList]
                        display(rawData.drop(columns=dropList)[intersectList])
                    else:    
                        display(self._filter_page(self.__data)) 
                
        def on_button_next(b):
            if self.__page<(self.__pages-1):
                set_header_and_buttons(self.__page+1) 
                button_previous.disabled=True
                button_next.disabled=True
                self._set_page(self.__page+1, records) 
                set_header_and_buttons(self.__page) 
                with data:
                    clear_output();
                    if not(fields==None):
                        rawData = self._filter_page(self.__data)
                        intersectList = [x for x in fields if x in rawData.columns.values]
                        dropList = [x for x in rawData.columns.values if x not in intersectList]
                        display(rawData.drop(columns=dropList)[intersectList])
                    else:    
                        display(self._filter_page(self.__data)) 
                
                
        set_header_and_buttons(self.__page)    
        with data:
            if not(fields==None):
                rawData = self._filter_page(self.__data)
                intersectList = [x for x in fields if x in rawData.columns.values]
                dropList = [x for x in rawData.columns.values if x not in intersectList]
                display(rawData.drop(columns=dropList)[intersectList])
            else:    
                display(self._filter_page(self.__data)) 
                  
        button_previous.on_click(on_button_previous)
        button_next.on_click(on_button_next)
        
        
        
        
        
        
        


from .pager import *
import json
import pandas as pd

class sample_type:

    def __init__(self, api, data):        
        if ((data is not None) & (type(data) == dict) & 
            ("sampleTypeID" in data.keys()) & 
            ("name" in data.keys())
           ):
            self.__api = api
            self.__sampleTypeID = data["sampleTypeID"]
            self.__name = data["name"]
            self.__title = "Sample type '"+self.__name+"' ("+str(self.__sampleTypeID)+")"
            self.__data = data
        else:
            raise Exception("no (valid) sampleType data") 
        
    def __repr__(self):
        return(self.__title)
        
    def name(self):
        """
        Get the name of the sample type.
        """
        return(self.__name)
        
    def id(self):
        """
        Get the id of the sample serie.
        """
        return(self.__sampleTypeID)    
        
    def data(self):
        return(self.__data)   
        
    def metas(self):
        return self.__api.sample_type_metas(self.__sampleTypeID)               
    
    


from .pager import *
from .experiment import *
from .section import *
from .group import *
from .sample_type import *
from .sample_type_meta import *
from .sample import *
from .sample_meta import *
from .sample_serie import *

import requests
import json
import keyring
import numbers

def reset_key():
    keyring.delete_password("elabjournal-python", "apikey")   

class api:

    def __init__(self, key=None):
        """
        Internal use only: initialize API
        """        
        self.__key = None
        self.__timeout = 180
        self.__url = "https://www.elabjournal.com"
        self.__user = None
        
        def check_and_set_key(key, throw_error, store_valid_key):
            if key==None:            
                return(False)
            elif (type(key) == str) & (len(key)>0):
                try:
                    rp = self.request("/api/v1/auth/user","get",{"body":{}},key)                   
                    if not(rp==None) & ("user" in rp.keys()):
                        self.__key = key
                        self.__user = rp["user"]
                        welcomeText = "Welcome "+rp["user"]["firstName"]+" "+rp["user"]["lastName"]                        
                        group = self.group_active()
                        if group is not None:
                            welcomeText = welcomeText + ", your active group is '" + str(group.name()) + "' (" + str(group.id()) + ")" 
                        print(welcomeText)
                        if store_valid_key:                                                
                            keyring.set_password("elabjournal-python", "apikey", key)                
                        return(True) 
                    else:
                        return(False)
                except:
                    return(False)
            else:
                return(False)               
        
        if key == None:
            keyring_key = keyring.get_password("elabjournal-python", "apikey")
            if not(check_and_set_key(keyring_key, False, False)):        
                from IPython.display import clear_output
                key = str(input("REST API key: "))
                clear_output();
                check_and_set_key(key, True, True)                                
        else:
          check_and_set_key(key, True, False)       
        
    
    def __repr__(self):
        """
        Internal use only: description API object
        """ 
        description = "eLABJournal API object"
        if self.__key == None:
            description += " - authentication failed"
        else:
            description += " - authenticated as "+self.__user["firstName"]+" "+self.__user["lastName"]
        return(description)    
            
    def samples(self, *args, **kwargs):
        """
        Get object to access samples.
        
        Parameters (object)
        ----------------------
        class: parser
            Pass the result of the first() method on this object instead
        class: sample_type
            Filter by sampleTypeID of this object 
        
        Parameters (key/value)
        ----------------------
        expand : str, optional
            Expand an ID field to an object
            separate values with comma for multiple expands
            location, quantity, meta, experiments, parent, children
        sort : str, optional    
            Sort by a specific field
        storageLayerID : str, optional
            Filter for samples in a storage layer
        sampleTypeID : str, optional
            Filter by sampleTypeID  
        name : str, optional
            Filter by sample name
        barcodes : str, optional
            Filter by barcodes (comma-separated)
        archived : str, optional
            Filter by archived or non-archived samples. 
            If set to true the search parameter has no effect
        quantityID : str, optional
            Filter by quantityID
        minimumQuantityAmount : str, optional
            Filter for samples that have a minimum quantity amount set
        checkedOut : str, optional
            Filter for checked out samples
        storageID : str, optional
            Filter for samples in a storage unit
        search : str, optional
            Search term to use for filtering samples. 
            This parameter can't be combined with archived=true
        """    
        request = {}
        kwargs_special = ["expand", "sort"]
        kwargs_keys = ["storageLayerID", "sampleTypeID", "name", "barcodes", 
                       "archived", "quantityID", "minimumQuantityAmount", "checkedOut", 
                       "storageID", "search"]
        if args is not None:
            for arg in args:
                check_arg = arg
                if type(check_arg)==pager:
                    check_arg = arg.first(True)
                if type(check_arg)==sample_type:
                    request["sampleTypeID"] = check_arg.id()
                else:
                    raise Exception("unsupported object '"+str(type(check_arg))+"'")                 
        if kwargs is not None:
            for key, value in kwargs.items():
                if key in kwargs_special:
                    request["$"+key] = value   
                elif key in kwargs_keys:
                    request[key] = value
                else:
                    raise Exception("unsupported key '"+key+"'")   
        return(pager(self, "Samples", "/api/v1/samples", request, "sampleID", None, 5, self.sample))
    
    def sample(self, id):
        """
        Get sample object with provided id (integer or string) or get multiple sample
        objects with provided id (list).
        """ 
        if isinstance(id,numbers.Integral) | isinstance(id,str): 
            rp = self.request("/api/v1/samples/"+str(id), "get", {})
            #check and get
            if (rp is not None) & (type(rp) == dict):
                return(sample(self,rp))                                                 
            else:
                return(None)    
        elif isinstance(id,list):
            request = {"sampleID": ",".join(map(str,id))}
            rp = self.request("/api/v1/samples/get/", "get", request)
            result = {}
            if (rp is not None) & isinstance(rp,list):
                for item in rp:
                    if isinstance(item,dict) & ("sampleID" in item.keys()):
                        result[item["sampleID"]] = sample(self,item)                           
            return(result)                                  
        else:
            raise Exception("incorrect call")         
                
        
    def sample_metas(self, id):
        """
        Get object to access sampleMetas for sample with provided id (integer or string).
        """ 
        request = {}
        if isinstance(id,numbers.Integral) | isinstance(id,str):
            sample = self.sample(id)
            if (sample is not None):                
                return(pager(self, "SampleMetas", "/api/v1/samples/"+str(id)+"/meta", request, "sampleMetaID", None, 5, sample.meta))
            else:
                return(None)    
        else:
            raise Exception("incorrect call")
                             
    def sample_meta(self, sample_id, sample_meta_id):
        """
        Get sample meta object for sample with provided sample_id (integer or string)
        and with provided sample_meta_id (integer or string) for the sample meta object.
        """ 
        if (isinstance(sample_id,numbers.Integral) | isinstance(sample_id,str)) & (isinstance(sample_meta_id,numbers.Integral) | isinstance(sample_meta_id,str)):
            rp = self.request("/api/v1/samples/"+str(sample_id)+"/meta/"+str(sample_meta_id), "get", {}) 
            #check and get
            if not(rp==None) & (type(rp) == dict):
                return(sample_meta(self,rp))                                                 
            else:
                return(None)                   
        else:
            raise Exception("incorrect call")
                             
    def sample_series(self):
        """
        Get object to access sampleSeries.
        """ 
        request = {}
        return(pager(self, "SampleSeries", "/api/v1/sampleSeries", request, "seriesID", None, 5, self.sample_serie))        
                             
    def sample_serie(self, id):
        """
        Get sample serie object with provided id (integer or string).
        """ 
        if isinstance(id,numbers.Integral) | isinstance(id,str):
            rp = self.request("/api/v1/sampleSeries/"+str(id), "get", {}) 
            #check and get
            if not(rp==None) & (type(rp) == dict):
                return(sample_serie(self,rp))                                                 
            else:
                return(None)                   
        else:
            raise Exception("incorrect call")
                             
    def samples_and_series(self, *args, **kwargs):
        """
        Get object to access samples and series in aggregated list.
        
        Parameters (object)
        ----------------------
        class: parser
            Pass the result of the first() method on this object instead
        class: sample_type
            Filter by sampleTypeID of this object 
        
        Parameters (key/value)
        ----------------------
        expand : str, optional
            Expand an ID field to an object
            separate values with comma for multiple expands
            location
        sort : str, optional    
            Sort by a specific field
        sampleTypeID : str, optional
            Filter by sampleTypeID  
        name : str, optional
            Filter by sample name
        search : str, optional
            Search term to use for filtering samples. 
        """    
        request = {}
        kwargs_special = ["expand", "sort"]
        kwargs_keys = ["expand", "sampleTypeID", "name", "search"]
        if args is not None:
            for arg in args:
                check_arg = arg
                if type(check_arg)==pager:
                    check_arg = arg.first(True)
                if type(check_arg)==sample_type:
                    request["sampleTypeID"] = check_arg.id()
                else:
                    raise Exception("unsupported object '"+str(type(check_arg))+"'")                 
        if kwargs is not None:
            for key, value in kwargs.items():
                if key in kwargs_special:
                    request["$"+key] = value   
                elif key in kwargs_keys:
                    request[key] = value
                else:
                    raise Exception("unsupported key '"+key+"'")    
        return(pager(self, "Samples and Series", "/api/v1/samplesAndSeries", request, ["type", "sampleID", "seriesID"], None, 5, self.sample_or_serie))
    
    def sample_or_serie(self, type, sample_id, sample_serie_id):
        """
        Get sample serie or serie object based on provided type (SAMPLE/SERIES)
        with provided sample_id (integer or string) or sample_serie_id (integer or string).
        """ 
        if type=="SERIES":
            return(self.sample_serie(sample_serie_id))
        elif type=="SAMPLE":
            return(self.sample(sample_id))
        else:
            raise Exception("incorrect call")
    
    def sample_types(self, *args, **kwargs):
        """
        Get object to access sampleTypes.
        
        Parameters (key/value)
        ----------------------
        expand : str, optional
            Expand an ID field to an object
        sort : str, optional    
            Sort by a specific field
        name : str, optional
            Filter by sample name
        archived : str, optional
            Filter by archived or non-archived samples.
        """    
        request = {}
        kwargs_special = ["expand", "sort"]
        kwargs_keys = ["name", "archived"]
        if args is not None:
            for arg in args:
                check_arg = arg
                if type(check_arg)==pager:
                    check_arg = arg.first(True)
                raise Exception("unsupported object '"+str(type(check_arg))+"'")                 
        if kwargs is not None:
            for key, value in kwargs.items():
                if key in kwargs_special:
                    request["$"+key] = value   
                elif key in kwargs_keys:
                    request[key] = value
                else:
                    raise Exception("unsupported key '"+key+"'")             
        return(pager(self, "SampleTypes", "/api/v1/sampleTypes", request, "sampleTypeID", None, 5, self.sample_type))
        
    def sample_type(self, id):
        """
        Get sampleType object with provided id (integer or string) or
        return the sampleType for a sample if a sample object is provided.
        """ 
        if isinstance(id,numbers.Integral) | isinstance(id,str):
            rp = self.request("/api/v1/sampleTypes/"+str(id), "get", {}) 
            #check and get
            if not(rp==None) & (type(rp) == dict):
                return(sample_type(self,rp))                                                 
            else:
                return(None)
        elif isinstance(id,sample):
            return id_object.sampleType()
        else:
            raise Exception("incorrect call")
            
    def sample_type_metas(self, id):
        """
        Get object to access sampleTypeMetas for sample with provided id (integer or string).
        """ 
        request = {}
        if isinstance(id,numbers.Integral) | isinstance(id,str):
            return(pager(self, "SampleTypeMetas", "/api/v1/sampleTypes/"+str(id)+"/meta", request, ["sampleTypeID", "sampleTypeMetaID"], None, 5, self.sample_type_meta))
        else:
            raise Exception("incorrect call")
                             
    def sample_type_meta(self, sample_type_id, sample_type_meta_id):
        """
        Get sampleTypeMeta object with provided sample_type_id (integer or string) and
        sample_type_meta_id (integer or string).
        """ 
        if (isinstance(sample_type_id,numbers.Integral) | isinstance(sample_type_id,str)) & (isinstance(sample_type_meta_id,numbers.Integral) | isinstance(sample_type_meta_id,str)):
            rp = self.request("/api/v1/sampleTypes/"+str(sample_type_id)+"/meta/"+str(sample_type_meta_id), "get", {}) 
            #check and get
            if not(rp==None) & (type(rp) == dict):
                return(sample_type_meta(self,rp))                                                 
            else:
                return(None)
        else:
            raise Exception("incorrect call")
                             
    def groups(self, *args, **kwargs):
        """
        Get all groups that you have joined.
        """
        request = {}
        return(pager(self, "Groups", "/api/v1/groups", request, "groupID", None, 5))
    
    def group_active(self):
        """
        Get the active group.
        """
        rp = self.request("/api/v1/groups/active", "get", {}) 
        #check and get
        if not(rp==None) & (type(rp) == dict):
            return(group(self,rp))                                                 
        else:
            return(None)
    
    def projects(self, *args, **kwargs):
        """
        Get object to access projects.
        
        Parameters (key/value)
        ----------------------
        expand : str, optional
            Expand an ID field to an object
        sort : str, optional    
            Sort by a specific field
        search : str, optional
            Search experiments by name or contents
        """    
        request = {}
        kwargs_special = ["expand", "sort"]
        kwargs_keys = ["search"]
        if args is not None:
            for arg in args:
                check_arg = arg
                if type(check_arg)==pager:
                    check_arg = arg.first(True)
                raise Exception("unsupported object '"+str(type(check_arg))+"'")                 
        if kwargs is not None:
            for key, value in kwargs.items():
                if key in kwargs_special:
                    request["$"+key] = value   
                elif key in kwargs_keys:
                    request[key] = value
                else:
                    raise Exception("unsupported key '"+key+"'")             
        return(pager(self, "Projects", "/api/v1/projects", request, "projectID", "projectID DESC", 5))
    
    def experiments(self, *args, **kwargs):
        """
        Get object to access experiments.
        
        Parameters (key/value)
        ----------------------
        expand : str, optional
            Expand an ID field to an object
        sort : str, optional    
            Sort by a specific field
        studyID : str, optional
            Filter by study    
        projectID : str, optional
            Filter by project   
        search : str, optional
            Search experiments by name or contents
        """    
        request = {}
        kwargs_special = ["expand", "sort"]
        kwargs_keys = ["studyID", "projectID", "search"]
        if args is not None:
            for arg in args:
                check_arg = arg
                if type(check_arg)==pager:
                    check_arg = arg.first(True)
                raise Exception("unsupported object '"+str(type(check_arg))+"'")                 
        if kwargs is not None:
            for key, value in kwargs.items():
                if key in kwargs_special:
                    request["$"+key] = value   
                elif key in kwargs_keys:
                    request[key] = value
                else:
                    raise Exception("unsupported key '"+key+"'")                   
        return(pager(self, "Experiments", "/api/v1/experiments", request, "experimentID", "experimentID DESC", 5, self.experiment))
    
    def studies(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self,"Studies", "/api/v1/studies", request, "studyID", "studyID DESC", 5))
    
    def storage_types(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self,"Storage Types", "/api/v1/storageTypes", request, "storageTypeID", "storageTypeID DESC", 5))
    
    def experiment(self, experiment_id):
        request = {"$expand": experiment_id}
        rp = self.request("/api/v1/experiments", "get", request)
        #check and get
        if not(rp==None) & (type(rp) == dict):
            if ("totalRecords" in rp.keys()) & (rp["totalRecords"]>=1):
                if "data" in rp.keys():
                    if(rp["totalRecords"]==1):
                        return(experiment(self, rp["data"][0]))
                    else:
                        for dataItem in rp["data"]:
                            if(dataItem["experimentID"]==experiment_id):
                                return(experiment(self, dataItem))
                        return(None)               
                else:   
                    return(None)
            else:
                return(None)                                                  
        else:
            return(None)
        
    def section(self, section_id):
        rp = self.request("/api/v1/experiments/sections/"+str(section_id), "get", {})
        #check and get
        if not(rp==None) & (type(rp) == dict):
            return(section(self,rp))                                                 
        else:
            return(None)    
    
    
    def request(self, location, method, request, key=None, show_messages=True, stream=False, headers=None):  
        if key==None:
            key = self.__key
        try: 
            request_headers = {"Accept": "application/json"}
            request_headers.update({"Authorization": key})    
            if not(headers==None):
                request_headers.update(headers)
            if method=="get":
                if type(request) == dict:
                    data = request
                else:
                    raise Exception("unsupported type of request") 
                response = requests.get(self.__url+location, params=data, timeout=self.__timeout, headers=request_headers, stream=stream)
            elif method=="post":    
                if type(request) == str:
                    request_headers.update({"Content-Type": "application/json"})
                    data = request
                else:
                    raise Exception("unsupported type of request") 
                response = requests.post(self.__url+location, data=data, timeout=self.__timeout, headers=request_headers, stream=stream)                
            elif method=="put":    
                if (type(request) == str) | (type(request) == bytes):
                    data = request
                else:
                    raise Exception("unsupported type of request") 
                response = requests.put(self.__url+location, data=data, timeout=self.__timeout, headers=request_headers, stream=stream)                
            else:
                raise Exception("unsupported method")   
            response.raise_for_status() 
            #no content
            if response.status_code == 204:
                return(None)
            elif stream:
                return(response.content)
            else:    
                try:
                    return(json.loads(response.text))
                except json.JSONDecodeError as decodeErr:
                    print ("JSON Decode Error:",decodeErr) 
                    raise Exception("no (valid) response from eLABJournal")            
        except requests.exceptions.HTTPError as httpErr: 
            if show_messages:
                print ("Http Error:",httpErr) 
            try:
                return(json.loads(response.text))
            except json.JSONDecodeError as decodeErr:
                return(None)           
        except requests.exceptions.ConnectionError as connErr: 
            if show_messages:
                print ("Error Connecting:",connErr) 
            return(None)
        except requests.exceptions.Timeout as timeOutErr: 
            if show_messages:
                print ("Timeout Error:",timeOutErr) 
            return(None)
        except requests.exceptions.RequestException as reqErr: 
            if show_messages:
                print ("Something Else:",reqErr) 
            return(None)
        raise Exception("no (valid) response from eLABJournal")     
        
        


from .pager import *
from .sample_type import *
import json
import pandas as pd

class sample:

    def __init__(self, api, data):  
        """
        Internal use only: initialize sample
        """
        if ((data is not None) & (type(data) == dict) & 
            ("sampleID" in data.keys()) & 
            ("name" in data.keys())
           ):
            self.__api = api
            self.__sampleID = data["sampleID"]
            self.__name = data["name"]
            self.__title = "Sample '"+self.__name+"' ("+str(self.__sampleID)+")"
            if ("sampleType" in data.keys()) & (type(data["sampleType"]) == dict) & ("sampleTypeID" in data["sampleType"]):
                self.__sampleTypeID = data["sampleType"]["sampleTypeID"]
            else:
                self.__sampleTypeID = None
            if "seriesID" in data.keys():
                self.__seriesID = data["seriesID"]
            else:
                self.__seriesID = None                    
            self.__data = data
        else:
            raise Exception("no (valid) sample data") 
        
    def __repr__(self):
        """
        Internal use only: description sample object
        """ 
        return(self.__title)
        
    def name(self):
        """
        Get the name of the sample.
        """
        return(self.__name)
        
    def id(self):
        """
        Get the id of the sample.
        """
        return(self.__sampleTypeID)
        
    def sample_type(self):
        """
        Get the sampleType of the sample as an object.
        """
        return self.__api.sample_type(self.__sampleTypeID)        
        
    def sample_serie(self):
        """
        Get the sampleSerie of the sample as an object.
        """
        return self.__api.sample_type(self.__sampleTypeID)        
        
    def data(self):
        """
        Get the data describing the sample.
        """
        return(self.__data)   
        
    def metas(self):
        """
        Get all the meta items (metas) for the sample.
        """
        return self.__api.sample_metas(self.__sampleID)    
        
    def meta(self, sample_meta_id):
        """
        Get meta item (meta) with provided sample_meta_id for the sample.
        """
        return self.__api.sample_meta(self.__sampleID, sample_meta_id)             
    
    


from .pager import *
import json
import pandas as pd
import numbers

class sample_serie:

    def __init__(self, api, data):        
        """
        Internal use only: initialize sample serie
        """
        if ((data is not None) & (type(data) == dict) & 
            ("seriesID" in data.keys()) & 
            ("name" in data.keys())
           ):
            self.__api = api
            self.__seriesID = data["seriesID"]
            self.__name = data["name"]
            self.__title = "Sample serie '"+self.__name+"' ("+str(self.__seriesID)+")"
            self.__data = data
        else:
            raise Exception("no (valid) sampleSerie data") 
        
    def __repr__(self):
        """
        Internal use only: description sample serie object
        """ 
        return(self.__title)
        
    def name(self):
        """
        Get the name of the sample serie.
        """
        return(self.__name)
        
    def id(self):
        """
        Get the id of the sample serie.
        """
        return(self.__seriesID)    
        
    def data(self):
        """
        Get the data describing the sample serie.
        """
        return({k:self.__data[k] for k in self.__data if k!="samples"})    
        
    def samples(self):
        """
        Get a dict with the samples for this sample serie.
        The sampleID is used as a key, the value is a sample object.
        """
        result = {}
        if "samples" in self.__data:
            samplesData = self.__data["samples"]
            if isinstance(samplesData, list):
                for sampleItem in samplesData:
                    if isinstance(sampleItem,dict) & ("sampleID" in sampleItem):
                        result[sampleItem["sampleID"]] = self.__api.sample(sampleItem["sampleID"])
                    elif isinstance(sampleItem,numbers.Integral) | isinstance(sampleItem,str):
                        result[sampleItem] = self.__api.sample(sampleItem)                             
        return result               
    
    